// Example code for homework 6.

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// global parameters for simulation
float end_seconds; // final time
const float frames_per_second=24;
int nx, ny, nz; // grid size
int iterations;
float buoyancy;

// global variables
int frame=0;
float *u, *v, *w, *pressure, *smoke; // primary variables
float *groundsmoke; // part of boundary conditions
float *unew, *vnew, *wnew, *smokenew; // used during solve
float *divergence; // used during solve

// functions
void read_parameters(const char *filename);
void init_simulation();
void write_smoke(const char *filename);
void advance_smoke(float dt);
void advect_everything(float dt);
void add_buoyancy(float dt);
void apply_boundary_conditions();
void find_divergence();
void solve_pressure();
void add_pressure_gradient();

// start of definitions
int main(int argc, char **argv)
{
   const char *directory=".";
   char filename[1000];

   if(argc<2 || argc>3){
      fprintf(stderr,"Need arguments: parameter-file [optional-directory]\n");
      return 1;
   }else if(argc==3)
      directory=argv[2];

   read_parameters(argv[1]);
   init_simulation();

   sprintf(filename, "%s/smoke%04d", directory, frame);
   write_smoke(filename);

   while(frame/frames_per_second < end_seconds){
      ++frame;
      advance_smoke(1.0/frames_per_second);
      sprintf(filename, "%s/smoke%04d", directory, frame);
      write_smoke(filename);
      fputc(frame%10 ? '+' : '#', stdout); fflush(stdout);
   }
   return 0;
}

void read_parameters(const char *filename)
{
   int i;
   FILE *fp=fopen(filename, "rt");
   fscanf(fp, "%f", &end_seconds);
   fscanf(fp, "%f", &buoyancy);
   fscanf(fp, "%d", &iterations);
   fscanf(fp, "%d %d %d", &nx, &ny, &nz);
   groundsmoke=(float*)calloc(nx*nz, sizeof(float));
   for(i=0; i<nx*nz; ++i)
      fscanf(fp, "%f", groundsmoke+i);
   fclose(fp);
}

void init_simulation()
{
   u=(float*)calloc((nx+1)*ny*nz, sizeof(float));
   v=(float*)calloc(nx*(ny+1)*nz, sizeof(float));
   w=(float*)calloc(nx*ny*(nz+1), sizeof(float));
   pressure=(float*)calloc(nx*ny*nz, sizeof(float));
   smoke=(float*)calloc(nx*ny*nz, sizeof(float));
   apply_boundary_conditions();

   // working variables
   unew=(float*)calloc((nx+1)*ny*nz, sizeof(float));
   vnew=(float*)calloc(nx*(ny+1)*nz, sizeof(float));
   wnew=(float*)calloc(nx*ny*(nz+1), sizeof(float));
   smokenew=(float*)calloc(nx*ny*nz, sizeof(float));
   divergence=(float*)calloc(nx*ny*nz, sizeof(float));
}

void write_smoke(const char *filename)
{
   int i;
   FILE *fp=fopen(filename, "wt");
   fprintf(fp, "%d %d %d\n", nx, ny, nz);
   for(i=0; i<nx*ny*nz; ++i){
      fprintf(fp, "%.5g", smoke[i]);
      fputc((i+1)%8 ? ' ' : '\n', fp);
   }
   fclose(fp);
}

void advance_smoke(float dt)
{
   advect_everything(dt);
   add_buoyancy(dt);
   apply_boundary_conditions();
   find_divergence();
   solve_pressure();
   add_pressure_gradient();
}

float interp1d(float qneg1, float q0, float q1, float q2, float x)
{
   // @@@ derive and fill in the limited Catmull-Rom 1d interpolation formula
   // arguments are values of q at points -1, 0, 1, and 2, and an x value between 0 and 1.
   // return the interpolated q evaluated at x. Note: it should be be between q0 and q1.
}

float interp(const float *q, float x, float y, float z, int qnx, int qny, int qnz)
{
   int i, j, k, ilo, jlo, klo, ihi, jhi, khi;
   float fracx, fracy, fracz;

   if(x<0)           { i=0; fracx=0; }
   else if(x>=qnx-1) { i=qnx-2; fracx=1; }
   else              { i=(int)x; fracx=x-i; }
   if(i==0) ilo=0; else ilo=i-1;
   if(i==qnx-2) ihi=i+1; else ihi=i+2;

   if(y<0)           { j=0; fracy=0; }
   else if(y>=qny-1) { j=qny-2; fracy=1; }
   else              { j=(int)y; fracy=y-j; }
   if(j==0) jlo=0; else jlo=j-1;
   if(j==qny-2) jhi=j+1; else jhi=j+2;

   if(z<0)           { k=0; fracz=0; }
   else if(z>=qnz-1) { k=qnz-2; fracz=1; }
   else              { k=(int)z; fracz=z-k; }
   if(k==0) klo=0; else klo=k-1;
   if(k==qnz-2) khi=k+1; else khi=k+2;

   #define Q(i,j,k) (q[((i)*qny+(j))*qnz+(k)])
   return interp1d(interp1d(interp1d(Q(ilo,jlo,klo),Q(i,jlo,klo),Q(i+1,jlo,klo),Q(ihi,jlo,klo),fracx),
                            interp1d(Q(ilo,j,klo),Q(i,j,klo),Q(i+1,j,klo),Q(ihi,j,klo),fracx),
                            interp1d(Q(ilo,j+1,klo),Q(i,j+1,klo),Q(i+1,j+1,klo),Q(ihi,j+1,klo),fracx),
                            interp1d(Q(ilo,jhi,klo),Q(i,jhi,klo),Q(i+1,jhi,klo),Q(ihi,jhi,klo),fracx), fracy),
                   interp1d(interp1d(Q(ilo,jlo,k),Q(i,jlo,k),Q(i+1,jlo,k),Q(ihi,jlo,k),fracx),
                            interp1d(Q(ilo,j,k),Q(i,j,k),Q(i+1,j,k),Q(ihi,j,k),fracx),
                            interp1d(Q(ilo,j+1,k),Q(i,j+1,k),Q(i+1,j+1,k),Q(ihi,j+1,k),fracx),
                            interp1d(Q(ilo,jhi,k),Q(i,jhi,k),Q(i+1,jhi,k),Q(ihi,jhi,k),fracx), fracy),
                   interp1d(interp1d(Q(ilo,jlo,k+1),Q(i,jlo,k+1),Q(i+1,jlo,k+1),Q(ihi,jlo,k+1),fracx),
                            interp1d(Q(ilo,j,k+1),Q(i,j,k+1),Q(i+1,j,k+1),Q(ihi,j,k+1),fracx),
                            interp1d(Q(ilo,j+1,k+1),Q(i,j+1,k+1),Q(i+1,j+1,k+1),Q(ihi,j+1,k+1),fracx),
                            interp1d(Q(ilo,jhi,k+1),Q(i,jhi,k+1),Q(i+1,jhi,k+1),Q(ihi,jhi,k+1),fracx), fracy),
                   interp1d(interp1d(Q(ilo,jlo,khi),Q(i,jlo,khi),Q(i+1,jlo,khi),Q(ihi,jlo,khi),fracx),
                            interp1d(Q(ilo,j,khi),Q(i,j,khi),Q(i+1,j,khi),Q(ihi,j,khi),fracx),
                            interp1d(Q(ilo,j+1,khi),Q(i,j+1,khi),Q(i+1,j+1,khi),Q(ihi,j+1,khi),fracx),
                            interp1d(Q(ilo,jhi,khi),Q(i,jhi,khi),Q(i+1,jhi,khi),Q(ihi,jhi,khi),fracx), fracy), fracz);
   /* this is the trilinear interpolation, for comparison
   return (1-fracx)*( (1-fracy)*( (1-fracz)*Q(i,j,k) + fracz*Q(i,j,k+1) )
                       + fracy*( (1-fracz)*Q(i,j+1,k) + fracz*Q(i,j+1,k+1) ) )
          + fracx*( (1-fracy)*( (1-fracz)*Q(i+1,j,k) + fracz*Q(i+1,j,k+1) )
                     + fracy*( (1-fracz)*Q(i+1,j+1,k) + fracz*Q(i+1,j+1,k+1) ) );
   */
   #undef Q
}

#define U(i,j,k) (u[((i)*ny+(j))*nz+(k)])
#define Unew(i,j,k) (unew[((i)*ny+(j))*nz+(k)])
#define V(i,j,k) (v[((i)*(ny+1)+(j))*nz+(k)])
#define Vnew(i,j,k) (vnew[((i)*(ny+1)+(j))*nz+(k)])
#define W(i,j,k) (w[((i)*ny+(j))*(nz+1)+(k)])
#define Wnew(i,j,k) (wnew[((i)*ny+(j))*(nz+1)+(k)])
#define S(i,j,k) (smoke[((i)*ny+(j))*nz+(k)])
#define Snew(i,j,k) (smokenew[((i)*ny+(j))*nz+(k)])
#define groundS(i,k) (groundsmoke[(i)*nz+(k)])
#define P(i,j,k) (pressure[((i)*ny+(j))*nz+(k)])
#define DIV(i,j,k) (divergence[((i)*ny+(j))*nz+(k)])

void advect_everything(float dt)
{
   int i, j, k;
   float uold, vold, wold, *swap;

   // @@@ advect u, v, w and s, storing the new values in unew, vnew, wnew, snew.

   // then swap so u,v,w,smoke contain the new advected values
   swap=u; u=unew; unew=swap;
   swap=v; v=vnew; vnew=swap;
   swap=w; w=wnew; wnew=swap;
   swap=smoke; smoke=smokenew; smokenew=swap;
}

void add_buoyancy(float dt)
{
   // @@@ add the buoyancy force to v
}

void apply_boundary_conditions()
{
   int i, k;
   for(i=0; i<nx; ++i){
      for(k=0; k<nz; ++k){
         V(i,0,k)=3*buoyancy*groundS(i,k);
         S(i,0,k)=groundS(i,k);
      }
   }
}

void find_divergence()
{
   // @@@ store the divergence of u,v,w in divergence
}

void solve_pressure()
{
   // @@@ use specified number of Gauss-Seidel iterations to solve for pressure
}

void add_pressure_gradient()
{
   // @@@ subtract the pressure gradient from the velocity field to make it divergence free
   //  (careful with signs - it may be more convenient to solve for -p above and add it here...)
}

